/* 
	v-m.s

	Volt-meter for RS-232 Interface
	Version 1
	By Uros Platise May 10, 1999 (c) Copyright
	
	Note: Default baudrate is 19200 bps
*/

#arch AT90S2313
#include "avr.inc"


/*-------------------*/
/* User Declarations */
/*===================*/
#define public DS_ADC_OFFSET_VOLTAGE	3
#define AVR_RET_REGISTER		r2

#define SYS_CLK				8000000
#define UART_BAUD_RATE			UART_UBBR(SYS_CLK, 19200)


/*------------------*/
/* V-m Declarations */
/*==================*/
#define __zero_reg__	r1
#define GPR		r16	/* general purpose register */
#define DATA		r17
#define DATAl		r18
#define DATAh		r19
#define DIV		r20

#define PRINT_CH(x)	\
	ldi	DATA, x \
        rcall	print_ch

#define PRINT_STR(ptr)	\
	ldi	ZL, low(ptr)	\
        ldi	ZH, high(ptr)	\
        rcall	print_str

#define PRINT_HEX(x)	\
	ldi	DATA, x	\
        rcall	print_hex
        

/*-----------------------*/
/* AVR Interrupt Vectors */
/*=======================*/
	seg abs=0 flash.ireq_vectors
        rjmp	__reset__
        reti
        reti
        reti
        reti
        reti
        reti
        reti
        reti
        reti
        reti


/*------*/
/* Main */
/*======*/
        seg flash.vmeter
__reset__:
	/* set stack memory and set zero register */
	ldi	GPR, vm_stack
        out	SPL, GPR
	clr	__zero_reg__

	/* Init RS-232 Port baudrate = 19200 */
        ldi	GPR, UART_BAUD_RATE
        out	UBRR, GPR
        ldi	GPR, BV(TXEN)
        out	UCR, GPR

        /* send test char */
        PRINT_STR(hello)

	/* Init ADC */
extern  _ds_adc_init
extern	_ds_adc     

	rcall	_ds_adc_init   

        /* Start Conversion */
convert:
	/* Call ADC and print result */
        rcall	_ds_adc
	PRINT_STR(volti)

	/* Adjust ADC result to 5.0 V scale */
        ldi	ZL,low(correction_tbl)
	ldi	ZH,high(correction_tbl)
	ldi	GPR,1
	eor	AVR_RET_REGISTER, GPR
	add	ZL, AVR_RET_REGISTER
	adc	ZH, __zero_reg__
	lpm
	mov	DATA,r0

	/* Print Decimal Number */
        rcall	print_dec
	PRINT_STR(enota)
	rjmp	convert


/*---------------------------------------------------*/
/* Print String: Z -> pointer to 0 terminated string */
/*===================================================*/
print_str:
	ldi	GPR, 1
next_ch:        
        eor	ZL, GPR
	lpm
        tst	r0
        breq	exit_pstr
wait_udre:        
        sbis	USR, UDRE
        rjmp	wait_udre
        out	UDR, r0
        eor	ZL, GPR
        adiw	ZL, 1
        rjmp	next_ch
exit_pstr:        
	ret

/*---------------------------*/
/* Print Character from DATA */
/*===========================*/
print_ch:
	sbis	USR, UDRE
        rjmp	print_ch
        out	UDR, DATA
	ret

/*----------------------------------------*/
/* Print Decimal Number and multiply by 2 */
/*========================================*/
print_dec:
	/* Multiply by 2 */
	mov	DATAl,DATA
	clr	DATAh
	clc	
	rol	DATAl
	rol	DATAh

	/* Print Decimal Result from 9 bits */
	ldi	DIV,100
	rcall	cntdnNprint	
	PRINT_CH(0x2e)		/* print full stop */
	ldi	DIV,10
	rcall	cntdnNprint
	ldi	DIV,1
	rcall	cntdnNprint	
	ret

/*-------------------------------------------*/
/* Count down and substract the value in DIV */
/*===========================================*/
cntdnNprint:
	ldi	DATA,0x30
start_sub:
	/* if (DATAh:DATAl < DIV) goto end_sub */
	cp	DATAl,DIV
	cpc	DATAh,__zero_reg__
	brlo	end_sub

	/* DATAh:DATAl =- DIV */
	sub	DATAl,DIV
	sbc	DATAh,__zero_reg__

	/* Increment output character and loop again */
	inc	DATA
	rjmp	start_sub
end_sub:
	rcall	print_ch
	ret

/*--------------*/
/* Stack Memory */
/*==============*/
	seg eram.stack
	ds.b	20
vm_stack:
	ds.b	1

/*----------*/
/* Messages */
/*==========*/
	seg flash.messages
hello:	dc.b	10,"Hello ADC",10,0
volti:	dc.b 	13, "U = ",0
enota:	dc.b	" V ",0


/*-------------------------------------------------------------*/
/* The following table produces 1 LSB error on four positions  */
/* due to 8 bit treatment. 9 bit table should be used for      */
/* exact translation.                                          */
/*                    U'/2 = adc_result*250/255 [V]            */
/*=============================================================*/
	seg flash.tbl8
correction_tbl:
	dc.b	0x00, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06,
		0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e,
		0x0f, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16,
		0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e,
		0x1f, 0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26,
		0x27, 0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e,
	 	0x2f, 0x30, 0x31, 0x32, 0x32, 0x33, 0x34, 0x35,
		0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d,
		0x3e, 0x3f, 0x40, 0x41, 0x42, 0x43, 0x44, 0x45,
		0x46, 0x47, 0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d,
		0x4e, 0x4f, 0x50, 0x51, 0x52, 0x53, 0x54, 0x55,
		0x56, 0x57, 0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d,
		0x5e, 0x5f, 0x60, 0x61, 0x62, 0x63, 0x64, 0x64,
		0x65, 0x66, 0x67, 0x68, 0x69, 0x6a, 0x6b, 0x6c,
		0x6d, 0x6e, 0x6f, 0x70, 0x71, 0x72, 0x73, 0x74,
		0x75, 0x76, 0x77, 0x78, 0x79, 0x7a, 0x7b, 0x7c,
		0x7d, 0x7e, 0x7f, 0x80, 0x81, 0x82, 0x83, 0x84,
		0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x8b, 0x8c,
		0x8d, 0x8e, 0x8f, 0x90, 0x91, 0x92, 0x93, 0x94, 
		0x95, 0x96, 0x96, 0x97, 0x98, 0x99, 0x9a, 0x9b, 
		0x9c, 0x9d, 0x9e, 0x9f, 0xa0, 0xa1, 0xa2, 0xa3, 
		0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xab, 
		0xac, 0xad, 0xae, 0xaf, 0xb0, 0xb1, 0xb2, 0xb3, 
		0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xbb, 
		0xbc, 0xbd, 0xbe, 0xbf, 0xc0, 0xc1, 0xc2, 0xc3, 
		0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc8, 0xc9, 0xca, 
		0xcb, 0xcc, 0xcd, 0xce, 0xcf, 0xd0, 0xd1, 0xd2, 
		0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 
		0xdb, 0xdc, 0xdd, 0xde, 0xdf, 0xe0, 0xe1, 0xe2, 
		0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea, 
		0xeb, 0xec, 0xed, 0xee, 0xef, 0xf0, 0xf1, 0xf2, 
		0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9, 0xfa

/* end of v-m.s */